/*
 * This file is part of the sn package.
 * Distribution of sn is covered by the GNU GPL. See file COPYING.
 * Copyright © 1998-2000 Harold Tay.
 * Copyright © 2000- Patrik Rådman.
 */

#ifndef PARAMETERS_H
#define PARAMETERS_H

extern char *snroot;
extern uid_t snuid;
extern gid_t sngid;

extern void parameters (bool wantwriteperms);

#endif
