/*
 * This file is part of the sn package.
 * Distribution of sn is covered by the GNU GPL. See file COPYING.
 * Copyright © 1998-2000 Harold Tay.
 * Copyright © 2000- Patrik Rådman.
 */

#ifndef DEFS_H
#define DEFS_H

#undef FALSE   /* Mac OS X apparently defines these */
#undef TRUE
typedef enum { FALSE, TRUE } bool;

#endif /* DEFS_H */
