/*
 * This file is part of the sn package.
 * Distribution of sn is covered by the GNU GPL. See file COPYING.
 * Copyright © 1998-2000 Harold Tay.
 * Copyright © 2000- Patrik Rådman.
 */

#ifndef VALID_H
#define VALID_H

extern bool is_valid_name (char *name);
extern bool is_valid_group (char *name);

#endif
